import dpdata
import numpy as np


def get_amber_fp(
    cutoff: float,
    target: str = ":1",
    out: str = None,
) -> dpdata.MultiSystems:
    """(qmwater-qmwater_not)-(mmwater-mmwater_not)
    = (qmwater-mmwater) - (qmwater_not-mmwater_not).
    """
    ms = dpdata.MultiSystems()
    ep = r"@%EP"
    if cutoff > 0.0:
        interactwith = f"({target})<@{cutoff:f}&!{ep}"
    else:
        interactwith = target

    s_qmwater = dpdata.LabeledSystem("qmwater", fmt="amber/md/qmmm", qm_region=target)
    s_qmwater = s_qmwater.remove_atom_names("EP")
    s_mmwater = dpdata.LabeledSystem("mmwater", fmt="amber/md/qmmm", qm_region=target)
    s_mmwater = s_mmwater.remove_atom_names("EP")
    assert np.all(s_qmwater["coords"] == s_mmwater["coords"])
    s_allcorr = s_mmwater.correction(s_qmwater)

    s_qmwaternot = dpdata.LabeledSystem("qmwater_not", fmt="amber/md/qmmm")
    s_qmwaternot = s_qmwaternot.remove_atom_names("EP")
    s_mmwaternot = dpdata.LabeledSystem("mmwater_not", fmt="amber/md/qmmm")
    s_mmwaternot = s_mmwaternot.remove_atom_names("EP")
    assert np.all(s_qmwaternot["coords"] == s_mmwaternot["coords"])
    s_notcorr = s_mmwaternot.correction(s_qmwaternot)

    # manually process
    natom_not = s_notcorr.get_natoms()
    assert np.all(s_allcorr["coords"][:, -natom_not:] == s_notcorr["coords"])
    s_allcorr.data["energies"] -= s_notcorr["energies"]
    s_allcorr.data["forces"][:, -natom_not:] -= s_notcorr["forces"]

    s_allcorr = s_allcorr.pick_by_amber_mask(
        "noepw.parm7", interactwith, pass_coords=True, nopbc=True
    )
    for ss in s_allcorr:
        ms.append(ss)

    if out:
        ms.to_deepmd_npy(out)
    return ms


def run(args):
    get_amber_fp(
        cutoff=args.cutoff,
        target=args.qm_region,
        out=args.out,
    )



import dpdata
import numpy as np
from ase.geometry import Cell, get_distances, wrap_positions
from dpdata.amber.mask import pick_by_amber_mask


def wrap_system(
    s_corr: dpdata.System, parmfile: str, target: str, cutoff: float
) -> dpdata.MultiSystems:
    """Wrap the coordinates around the QM region.

    Parameters
    ----------
    s_corr: dpdata.System
        The system to be processed.
    parmfile: str
        The parm7 file.
    target: str
        The QM system mask.
    cutoff: float
        The QM/MM cutoff radius.

    Returns
    -------
    ms: dpdata.MultiSystems
        The output MultiSystems
    """
    ms = dpdata.MultiSystems()
    ep = r"@%EP"
    if cutoff > 0.0:
        interactwith = f"({target})<@{cutoff:f}&!{ep}"
    else:
        interactwith = target
    # wrap the coords...
    qm_index = pick_by_amber_mask(parmfile, target)
    qm_coords = s_corr["coords"][:, qm_index, :]
    for ii in range(len(s_corr)):
        cell = Cell(s_corr["cells"][ii])
        qm_coord = qm_coords[ii]
        qm_distances = get_distances(qm_coord, cell=cell, pbc=True)[1]
        # find the coord that has the minimal total distance as the center
        center = qm_coords[ii, np.argmin(np.sum(qm_distances, axis=1))]
        wraped_coords = wrap_positions(
            s_corr["coords"][ii],
            cell=s_corr["cells"][ii],
            pbc=True,
            center=cell.scaled_positions(center),
        )
        s_corr["coords"][ii, :, :] = wraped_coords

    s_corr = s_corr.pick_by_amber_mask(
        parmfile, interactwith, pass_coords=True, nopbc=True
    )
    for ss in s_corr:
        if "EP" in ss["atom_names"]:
            ss = ss.remove_atom_names("EP")
        ms.append(ss)
    return ms



import os
import re

import numpy as np
from ase.geometry import cellpar_to_cell
from dpdata.amber.mask import pick_by_amber_mask
from dpdata.format import Format
from dpdata.periodic_table import ELEMENTS
from dpdata.plugins.amber import AmberMDFormat
from dpdata.unit import EnergyConversion
from scipy.io import netcdf

kcalmol2eV = EnergyConversion("kcal_mol", "eV").value()
symbols = ["X"] + ELEMENTS

energy_convert = kcalmol2eV
force_convert = energy_convert


def read_amber_traj(
    parm7_file,
    nc_file,
    mdfrc_file=None,
    mden_file=None,
    mdout_file=None,
    qm_region=None,
    labeled=True,
    exclude_unconverged=True,
):
    """Read amber trajectory.

    The amber trajectory includes:
    * nc, NetCDF format, stores coordinates
    * mdfrc, NetCDF format, stores forces
    * mden (optional), text format, stores energies
    * mdout (optional), text format, may store energies if there is no mden_file
    * parm7, text format, stores types.

    Parameters
    ----------
    parm7_file, nc_file, mdfrc_file, mden_file, mdout_file: str
        filenames
    qm_region: None or list or str
        amber mask
    labeled: bool
        the output is a dpdata.LabeledSystem or dpdata.System
    exclude_unconverged: bool
        exclude unconverged frames
    """
    flag_atom_type = False
    flag_atom_numb = False
    amber_types = []
    atomic_number = []
    with open(parm7_file) as f:
        for line in f:
            if line.startswith("%FLAG"):
                flag_atom_type = line.startswith("%FLAG AMBER_ATOM_TYPE")
                flag_atom_numb = line.startswith("%FLAG ATOMIC_NUMBER")
            elif flag_atom_type or flag_atom_numb:
                if line.startswith("%FORMAT"):
                    fmt = re.findall(r"\d+", line)
                    fmt0 = int(fmt[0])
                    fmt1 = int(fmt[1])
                else:
                    for ii in range(fmt0):
                        start_index = ii * fmt1
                        end_index = (ii + 1) * fmt1
                        if end_index >= len(line):
                            continue
                        content = line[start_index:end_index].strip()
                        if flag_atom_type:
                            amber_types.append(content)
                        elif flag_atom_numb:
                            atomic_number.append(int(content))
    ml_types = []
    if qm_region is None:
        qm_region = []

    if isinstance(qm_region, str):
        qm_region = pick_by_amber_mask(parm7_file, qm_region)

    for ii, (tt, nn) in enumerate(zip(amber_types, atomic_number)):
        if ii in qm_region:
            ml_types.append(symbols[nn])
        elif tt in ("OW", "HW", "EP"):
            ml_types.append(tt)
        else:
            ml_types.append("m" + symbols[nn])

    with netcdf.netcdf_file(nc_file, "r") as f:
        coords = np.array(f.variables["coordinates"][:])
        shape = coords.shape
        cells = np.zeros((shape[0], 3, 3))
        if "cell_lengths" in f.variables:
            nopbc = False
            cell_lengths = np.array(f.variables["cell_lengths"][:])
            cell_angles = np.array(f.variables["cell_angles"][:])
            for ii in range(cell_lengths.shape[0]):
                cells[ii, :, :] = cellpar_to_cell([*cell_lengths[ii], *cell_angles[ii]])
        else:
            nopbc = True

    if labeled:
        with netcdf.netcdf_file(mdfrc_file, "r") as f:
            forces = np.array(f.variables["forces"][:])

        # load energy from mden_file or mdout_file
        energies = []
        if mdout_file is not None and os.path.isfile(mdout_file):
            is_coverage = True
            with open(mdout_file) as f:
                for line in f:
                    if (
                        line.strip()
                        == "QMMM SCC-DFTB: Convergence could not be achieved in this step."
                    ):
                        is_coverage = False
                    elif line.strip().startswith("A V E R A G E S   O V E R"):
                        # end of MD simulation
                        break
                    elif "EPtot" in line:
                        if is_coverage:
                            s = line.split()
                            energies.append(float(s[-1]))
                        else:
                            energies.append(np.nan)
                            is_coverage = True
        elif mden_file is not None and os.path.isfile(mden_file):
            with open(mden_file) as f:
                for line in f:
                    if line.startswith("L6"):
                        s = line.split()
                        if s[2] != "E_pot":
                            energies.append(float(s[2]))
        else:
            raise RuntimeError("Please provide one of mden_file and mdout_file")

    atom_names, atom_types, atom_numbs = np.unique(
        ml_types, return_inverse=True, return_counts=True
    )

    data = {}
    data["atom_names"] = list(atom_names)
    data["atom_numbs"] = list(atom_numbs)
    data["atom_types"] = atom_types
    data["coords"] = coords
    data["cells"] = cells
    if labeled:
        data["forces"] = forces * force_convert
        data["energies"] = np.array(energies) * energy_convert
        if exclude_unconverged:
            # exclude nan index
            idx_pick = ~np.logical_or(
                np.isnan(data["energies"]), np.isnan(data["forces"]).any(axis=(1, 2))
            )
            data["coords"] = data["coords"][idx_pick, :, :]
            data["cells"] = data["cells"][idx_pick, :, :]
            data["energies"] = data["energies"][idx_pick]
            data["forces"] = data["forces"][idx_pick, :, :]
    data["orig"] = np.array([0, 0, 0])
    if nopbc:
        data["nopbc"] = True
    return data


@Format.register("amber/md/qmmm")
class AmberMDQMMMFormat(AmberMDFormat):
    def from_system(
        self, file_name=None, parm7_file=None, nc_file=None, qm_region=None, **kwargs
    ):
        # assume the prefix is the same if the spefic name is not given
        if parm7_file is None:
            parm7_file = file_name + ".parm7"
        if nc_file is None:
            nc_file = file_name + ".nc"
        return read_amber_traj(
            parm7_file=parm7_file, nc_file=nc_file, qm_region=qm_region, labeled=False
        )

    def from_labeled_system(
        self,
        file_name=None,
        parm7_file=None,
        nc_file=None,
        mdfrc_file=None,
        mden_file=None,
        mdout_file=None,
        qm_region=None,
        exclude_unconverged=True,
        **kwargs,
    ):
        # assume the prefix is the same if the spefic name is not given
        if parm7_file is None:
            parm7_file = file_name + ".parm7"
        if nc_file is None:
            nc_file = file_name + ".nc"
        if mdfrc_file is None:
            mdfrc_file = file_name + ".mdfrc"
        if mden_file is None:
            mden_file = file_name + ".mden"
        if mdout_file is None:
            mdout_file = file_name + ".mdout"
        return read_amber_traj(
            parm7_file,
            nc_file,
            mdfrc_file,
            mden_file,
            mdout_file,
            qm_region,
            exclude_unconverged=exclude_unconverged,
        )



import argparse

from .corr import run as corr_run
from .model_devi import run as model_devi_run
from .qmbuffer import run as qmwater_run
from .qmwater_corr import run as qmwater_corr_run


def run():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    parser_model_devi = subparsers.add_parser(
        "devi", help="Calculate model deviations."
    )
    parser_model_devi.add_argument("--models", type=str, nargs="+", help="Models")
    parser_model_devi.add_argument("--cutoff", type=float, help="QM/MM cutoff radius")
    parser_model_devi.add_argument(
        "--parm7_file", type=str, default="qmmm.parm7", help="AMBER parm file"
    )
    parser_model_devi.add_argument(
        "--qm_region",
        type=str,
        default=":1",
        help="QM region with AMBER mask. Add quote if containing space",
    )
    parser_model_devi.add_argument("--prefix", type=str, help="prefix to the .nc file")
    parser_model_devi.set_defaults(func=model_devi_run)

    parser_corr = subparsers.add_parser("corr", help="Generate systems for DPRc.")
    parser_corr.add_argument("--cutoff", type=float, help="QM/MM cutoff radius")
    parser_corr.add_argument(
        "--parm7_file", type=str, default="qmmm.parm7", help="AMBER parm file"
    )
    parser_corr.add_argument(
        "--qm_region",
        type=str,
        default=":1",
        help="QM region with AMBER mask. Add quote if containing space",
    )
    parser_corr.add_argument("--nc", type=str, help="AMBER coordinates (nc) file")
    parser_corr.add_argument(
        "--hl", type=str, help="Prefix to high-level files, including .mdout, .mdfrc"
    )
    parser_corr.add_argument(
        "--ll", type=str, help="Prefix to low-level files, including .mdout, .mdfrc"
    )
    parser_corr.add_argument(
        "--out",
        type=str,
        default="dataset",
        help="output directory or hdf5 file for DeePMD-kit data (default is dataset)",
    )
    parser_corr.add_argument("--suffix_mdfrc", type=str, help="suffix of mdfrc file")
    parser_corr.set_defaults(func=corr_run)

    parser_qmwater = subparsers.add_parser(
        "qmwater", help="Generate QM water parm and mdin file."
    )
    parser_qmwater.add_argument("--cutoff", type=float, help="cutoff")
    parser_qmwater.add_argument(
        "--parm7_file", type=str, default="qmmm.parm7", help="parm7_file"
    )
    parser_qmwater.add_argument("--qm_region", type=str, default=":1", help="qm_region")
    parser_qmwater.add_argument("--nc", type=str, help="nc file")
    parser_qmwater.add_argument("--hl", type=str, help="high level mdin file")
    parser_qmwater.add_argument("--ll", type=str, help="low level mdin file")
    parser_qmwater.add_argument(
        "--charge", type=str, default="dataset", help="Charge of QM region"
    )
    parser_qmwater.set_defaults(func=qmwater_run)

    parser_qmwater_corr = subparsers.add_parser(
        "qmwater_corr", help="Generate systems for QM water."
    )
    parser_qmwater_corr.add_argument("--cutoff", type=float, help="cutoff")
    parser_qmwater_corr.add_argument(
        "--qm_region", type=str, default=":1", help="qm_region"
    )
    parser_qmwater_corr.add_argument(
        "--out",
        type=str,
        default="dataset",
        help="output directory (default is dataset)",
    )
    parser_qmwater_corr.set_defaults(func=qmwater_corr_run)

    args = parser.parse_args()
    args.func(args)



"""Calculate model deviation of the Amber MD traj."""
import dpdata
import numpy as np
from dpdata.amber.mask import load_param_file
from tqdm import tqdm


def calculate_devi(
    models: list,
    cutoff: float,
    parm7_file: str = "qmmm.parm7",
    qm_region: str = ":1",
    prefix: str = "rc",
):
    """Calculate model deviation of the Amber MD traj.

    Parameters
    ----------
    models: list[str]
        models, for example ["graph.0.pb", "graph.1.pb"]
    cutoff: float
        The QM/MM cutoff radius.
    parm7_file: str
        The parm7 file.
    qm_region: str
        The QM system mask.
    prefix: str
        The prefix of MD files.
    """
    # Deep potentials
    from deepmd.infer import DeepPot, calc_model_devi

    dps = [DeepPot(mm) for mm in models]

    sy = dpdata.System(
        file_name=prefix,
        parm7_file=parm7_file,
        fmt="amber/md/qmmm",
        qm_region=qm_region,
    )
    parm7 = load_param_file(parm7_file)
    ep = r"@%EP"
    if cutoff > 0.0:
        interactwith = f"({qm_region})<@{cutoff:f}&!{ep}"
    else:
        interactwith = qm_region

    stds = []
    for ii, ss in enumerate(tqdm(sy)):
        ms = ss.pick_by_amber_mask(parm7, interactwith, pass_coords=True)
        ss = list(ms.systems.values())[0]
        ss = ss.remove_atom_names("EP")
        devi = calc_model_devi(ss["coords"], ss["cells"], ss["atom_types"], dps)
        devi[0, 0] = ii
        stds.append(devi)
    stds = np.concatenate(stds)
    np.savetxt(
        "model_devi.out",
        stds,
        header="step max_devi_v min_devi_v avg_devi_v max_devi_f min_devi_f avg_devi_f devi_e",
    )


def run(args):
    calculate_devi(
        models=args.models,
        cutoff=args.cutoff,
        parm7_file=args.parm7_file,
        qm_region=args.qm_region,
        prefix=args.prefix,
    )



import shutil
from pathlib import Path

import dpdata
import numpy as np
import pytest

from dpamber.corr import get_amber_fp


@pytest.fixture
def test_system() -> dpdata.MultiSystems:
    """Tested system."""
    return dpdata.MultiSystems().from_deepmd_raw(
        Path(__file__).parent / "corr" / "dataset"
    )


def system_is_equal(sys1: dpdata.LabeledSystem, sys2: dpdata.LabeledSystem):
    """Check if two systems are equal.

    Parameters
    ----------
    sys1: dpdata.LabeledSystem
        The first system.
    sys2: dpdata.LabeledSystem
        The second system.

    Raises
    ------
    AssertionError
        If the two systems are not equal.
    """
    assert sys1["atom_names"] == sys2["atom_names"]
    assert sys1["atom_numbs"] == sys2["atom_numbs"]
    assert sys1.nopbc == sys2.nopbc
    np.testing.assert_almost_equal(sys1["atom_types"], sys2["atom_types"])
    np.testing.assert_almost_equal(sys1["coords"], sys2["coords"])
    np.testing.assert_almost_equal(sys1["cells"], sys2["cells"])
    np.testing.assert_almost_equal(sys1["energies"], sys2["energies"])
    np.testing.assert_almost_equal(sys1["forces"], sys2["forces"])


def get_single_system(multi: dpdata.MultiSystems) -> dpdata.LabeledSystem:
    """Assume the multi-system contains only one system.

    Parameters
    ----------
    multi: dpdata.MultiSystems
        The multi-system.

    Returns
    -------
    dpdata.LabeledSystem
        The system.
    """
    assert len(multi.systems) == 1
    return list(multi.systems.values())[0]


def test_corr(test_system):
    get_amber_fp(
        cutoff=6.0,
        parmfile=str(Path(__file__).parent / "corr/qmmm.parm7"),
        ncfile=str(Path(__file__).parent / "corr/rc.nc"),
        ll=str(Path(__file__).parent / "corr/high_level"),
        hl=str(Path(__file__).parent / "corr/low_level"),
        target=":1",
        out="tmp_data",
    )
    tmp_system = dpdata.MultiSystems().from_deepmd_npy("tmp_data")
    tmp_system_single = get_single_system(tmp_system)
    test_system_single = get_single_system(test_system)
    system_is_equal(tmp_system_single, test_system_single)


def test_corr_hdf5(test_system):
    get_amber_fp(
        cutoff=6.0,
        parmfile=str(Path(__file__).parent / "corr/qmmm.parm7"),
        ncfile=str(Path(__file__).parent / "corr/rc.nc"),
        ll=str(Path(__file__).parent / "corr/high_level"),
        hl=str(Path(__file__).parent / "corr/low_level"),
        target=":1",
        out="tmp_data.hdf5",
    )
    tmp_system = dpdata.MultiSystems().from_deepmd_hdf5("tmp_data.hdf5")
    tmp_system_single = get_single_system(tmp_system)
    test_system_single = get_single_system(test_system)
    system_is_equal(tmp_system_single, test_system_single)


def test_md_corr(test_system):
    shutil.copyfile(
        Path(__file__).parent / "corr/low_level.mdfrc",
        Path(__file__).parent / "md/md.mdfrc",
    )
    get_amber_fp(
        cutoff=6.0,
        parmfile=str(Path(__file__).parent / "corr/qmmm.parm7"),
        ncfile=str(Path(__file__).parent / "corr/rc.nc"),
        ll=str(Path(__file__).parent / "corr/high_level"),
        hl=str(Path(__file__).parent / "md/md"),
        target=":1",
        out="tmp_data",
    )
    tmp_system = dpdata.MultiSystems().from_deepmd_npy("tmp_data")
    tmp_system_single = get_single_system(tmp_system)
    test_system_single = get_single_system(test_system)
    system_is_equal(tmp_system_single, test_system_single)


def test_uncoverage_system():
    tmp_system = dpdata.LabeledSystem(
        str(Path(__file__).parent / "uncoverage/uncoverage"),
        nc_file=str(Path(__file__).parent / "corr/rc.nc"),
        mdfrc_file=str(Path(__file__).parent / "corr/low_level.mdfrc"),
        parm7_file=str(Path(__file__).parent / "corr/qmmm.parm7"),
        fmt="amber/md/qmmm",
        qm_region=":1",
    )
    assert len(tmp_system) == 0


def test_uncoverage_corr():
    shutil.copyfile(
        Path(__file__).parent / "corr/low_level.mdfrc",
        Path(__file__).parent / "uncoverage/uncoverage.mdfrc",
    )
    tmp_system = get_amber_fp(
        cutoff=6.0,
        parmfile=str(Path(__file__).parent / "corr/qmmm.parm7"),
        ncfile=str(Path(__file__).parent / "corr/rc.nc"),
        ll=str(Path(__file__).parent / "corr/high_level"),
        hl=str(Path(__file__).parent / "uncoverage/uncoverage"),
        target=":1",
    )
    assert len(tmp_system) == 0



from typing import Union

import dpdata
import numpy as np
from ase.geometry import Cell, get_distances, wrap_positions
from dpdata.amber.mask import pick_by_amber_mask


def get_amber_fp(
    cutoff: float,
    parmfile: str,
    ncfile: str,
    ll: str,
    hl: str,
    target: str = ":1",
    out: str = None,
    idx: Union[slice, list, int] = None,
    suffix_mdfrc=None,
) -> dpdata.MultiSystems:
    """Use Ambertools to do correction calculation between a high level potential and a low level potential.

    Parameters
    ----------
    cutoff: float
        The QM/MM cutoff radius.
    parmfile: str
        The original parm file.
    ncfile: str
        The coordinates file.
    ll: str
        The low level system prefix.
    hl: str
        The high level system prefix.
    target: str
        The QM system mask.
    out: str
        The output deepmd/npy directory.
    idx: slice or list or int
        index
    suffix_mdfrc: str, optional
        suffix of mdfrc file

    Returns
    -------
    ms: dpdata.MultiSystems
        The output MultiSystems
    """
    ms = dpdata.MultiSystems()
    ep = r"@%EP"
    if cutoff > 0.0:
        interactwith = f"({target})<@{cutoff:f}&!{ep}"
    else:
        interactwith = target

    if suffix_mdfrc is not None:
        ll_frc = ll + suffix_mdfrc
        hl_frc = hl + suffix_mdfrc
    else:
        ll_frc = None
        hl_frc = None

    s_ll = dpdata.LabeledSystem(
        ll,
        nc_file=ncfile,
        mdfrc_file=ll_frc,
        parm7_file=parmfile,
        fmt="amber/md/qmmm",
        qm_region=target,
        exclude_unconverged=False,
    )
    s_hl = dpdata.LabeledSystem(
        hl,
        nc_file=ncfile,
        mdfrc_file=hl_frc,
        parm7_file=parmfile,
        fmt="amber/md/qmmm",
        qm_region=target,
        exclude_unconverged=False,
    )
    if idx is not None:
        s_ll = s_ll[idx]
        s_hl = s_hl[idx]

    s_corr = s_ll.correction(s_hl)

    # remove unconverged frames
    idx_pick = ~np.logical_or(
        np.isnan(s_corr["energies"]), np.isnan(s_corr["forces"]).any(axis=(1, 2))
    )
    s_corr = s_corr[idx_pick]

    # wrap the coords...
    qm_index = pick_by_amber_mask(parmfile, target)
    qm_coords = s_corr["coords"][:, qm_index, :]
    for ii in range(len(s_corr)):
        cell = Cell(s_corr["cells"][ii])
        qm_coord = qm_coords[ii]
        qm_distances = get_distances(qm_coord, cell=cell, pbc=True)[1]
        # find the coord that has the minimal total distance as the center
        center = qm_coords[ii, np.argmin(np.sum(qm_distances, axis=1))]
        wraped_coords = wrap_positions(
            s_corr["coords"][ii],
            cell=s_corr["cells"][ii],
            pbc=True,
            center=cell.scaled_positions(center),
        )
        s_corr["coords"][ii, :, :] = wraped_coords

    s_corr = s_corr.pick_by_amber_mask(
        parmfile, interactwith, pass_coords=True, nopbc=True
    )
    for ss in s_corr:
        if "EP" in ss["atom_names"]:
            ss = ss.remove_atom_names("EP")
        ms.append(ss)

    if out:
        if out.endswith(".hdf5"):
            ms.to_deepmd_hdf5(out)
        else:
            ms.to_deepmd_npy(out)
    return ms


def run(args):
    get_amber_fp(
        cutoff=args.cutoff,
        parmfile=args.parm7_file,
        ncfile=args.nc,
        ll=args.ll,
        hl=args.hl,
        target=args.qm_region,
        out=args.out,
        suffix_mdfrc=args.suffix_mdfrc,
    )



from pathlib import Path

import numpy as np

from dpamber.model_devi import calculate_devi


def test_model_devi(mocker):
    class FakeDP:
        def __init__(self, model) -> None:
            pass

        def eval(self, coords, cells, atom_types, mixed_type=False):
            return 0, np.zeros_like(coords), np.zeros(9)

    mocker.patch("deepmd.infer.DeepPot", FakeDP)

    # spy.assert_called_once_with(21)

    calculate_devi(
        ["graph.0.pb", "graph.1.pb"],
        6,
        str(Path(__file__).parent / "corr/qmmm.parm7"),
        ":1",
        str(Path(__file__).parent / "corr/rc"),
    )
    x = np.loadtxt("model_devi.out", ndmin=2)
    assert x.shape == (1, 8)



from copy import copy

import dpdata
import parmed.tools as PT
from dpdata.amber.mask import load_param_file, pick_by_amber_mask


def qmbuffer(
    cutoff: float,
    ncfile: str,
    parmfile: str,
    hl_mdinfile: str,
    ll_mdinfile: str,
    target: str = ":1",
    targetcharge: int = -2,
):
    """Assume there is only one frame.
    cutoff = 4.5.
    """
    mask_str = f"(({target})<:{cutoff:.1f}) & !({target})"
    s = dpdata.System(
        "",
        nc_file=ncfile,
        parm7_file=parmfile,
        fmt="amber/md",
        use_element_symbols=target,
    )
    # systems with nearby waters
    parm = load_param_file(parmfile)
    target_idx = pick_by_amber_mask(parm, target, s["coords"][0])
    nearby_idx = pick_by_amber_mask(parm, mask_str, s["coords"][0])
    parm = reorder(parm, target_idx, nearby_idx)
    # Here we need to generate 4 new parms:
    # 1. reordered, QM water
    # 2. reordered, QM water, no target
    # 3. reordered, MM water
    # 4. reordered, MM water, no target
    # parm 1
    new_nearby_idx = range(len(target_idx), len(target_idx) + len(nearby_idx))
    parm_qm = copy(parm)
    parm_qm, nwater = remove_epw(parm_qm, new_nearby_idx)
    water_mask = "(:2-%d)" % (nwater + 1)
    qmmask = f"({target})|({water_mask})"
    sort_parm(parm_qm)
    write_parm(parm_qm, "qmwater", hl_mdinfile, qmmask, target, charge=targetcharge)

    # parm2
    parm_qm_not = strip_atoms(parm_qm, target)
    water_mask = "(:1-%d)" % (nwater)
    sort_parm(parm_qm_not)
    write_parm(parm_qm_not, "qmwater_not", hl_mdinfile, water_mask, "")

    # parm3
    parm_mm = copy(parm)
    sort_parm(parm_mm)
    write_parm(parm_mm, "mmwater", ll_mdinfile, target, target, charge=targetcharge)

    # parm4
    parm_mm_not = strip_atoms(parm_mm, target)
    sort_parm(parm_mm_not)
    write_parm(parm_mm_not, "mmwater_not", ll_mdinfile, "", "")

    # also try to remove all epw
    parm_noep = copy(parm)
    parm_noep, _ = remove_epw(parm_noep, range(len(target_idx), s.get_natoms()))
    sort_parm(parm_mm)
    parm_noep.write_parm("noepw.parm7")


def label_atoms(parm, target_idx, qmwater_idx):
    """Set qwt."""
    for ii in range(len(parm.atoms)):
        atom = parm.atoms[ii]
        if ii in target_idx:
            atom._sorted = 0
        elif ii in qmwater_idx:
            atom._sorted = 1
        else:
            atom._sorted = 2


def remove_epw(parm, idx):
    """Remove all EP atoms in water in idx."""
    nwater = 0
    ep = []
    for ii in idx:
        atom = parm.atoms[ii]
        res = atom.residue
        if atom.type == "EP":
            ep.append(ii)
            nwater += 1
            if res.atoms[0].name == "O":
                res.name = "qwt"
                res.atoms[0].charge += atom.charge
                # print(atom.charge, res.atoms[0].charge)
                atom.charge = 0
                # remove from res
                res.atoms.pop(res.atoms.index(atom))
            else:
                raise RuntimeError()
            # remove bonds
            for bond in atom.bonds:
                bond.delete()
            for _ in range(len(atom._bond_partners)):
                atom._bond_partners.pop()
            for _ in range(len(atom._exclusion_partners)):
                atom._exclusion_partners.pop()
        else:
            for p in atom.bonds:
                if p.atom1.type == "EP" or p.atom2.type == "EP":
                    p.delete()
            for _ in range(len(atom._exclusion_partners)):
                atom._exclusion_partners.pop()
            for p in atom.children:
                if p.type == "EP":
                    atom.children.pop(atom.children.index(p))

    mask = "@" + ",".join([str(ii + 1) for ii in ep])
    parm = strip_atoms(parm, mask)
    return parm, nwater


def strip_atoms(parm, mask):
    parm.strip(mask)
    return parm


def reorder(parm, target_idx, nearby_idx):
    label_atoms(parm, target_idx, qmwater_idx=nearby_idx)
    parm.atoms.sort(key=lambda a: a._sorted)
    return parm


def sort_parm(parm):
    # sort residues
    parm.residues.sort(key=sort_residues)


def write_parm(parm, fn, mdinfile, qmmask, target, charge=0):
    parm.write_parm("%s.parm7" % fn)
    parm.write_rst7("%s.rst7" % fn)
    PT.writeCoordinates(parm, "%s.nc" % fn).execute()
    ifqnt = str(int(target != ""))
    with open(mdinfile) as f, open("%s.mdin" % fn, "w") as fw:
        fw.write(
            f.read()
            .replace("%QMMASK%", qmmask)
            .replace("%NOSHAKEMASK%", target)
            .replace("%CHARGE%", str(charge))
            .replace("%IFQNT%", ifqnt)
        )


def sort_residues(res):
    keys = {
        "qwt": 1,
        "WAT": 2,
    }
    return keys.get(res.name, 0)


def run(args):
    qmbuffer(
        cutoff=args.cutoff,
        parmfile=args.parm7_file,
        ncfile=args.nc,
        ll_mdinfile=args.ll,
        hl_mdinfile=args.hl,
        target=args.qm_region,
        targetcharge=args.charge,
    )
